/* datei.cc
 * This file belongs to Worker, a filemanager for UNIX/X11.
 * Copyright (C) 2001 Ralf Hoffmann.
 * You can contact me at: ralf.hoffmann@epost.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "datei.h"
#include <aguix/lowlevelfunc.h>
#include "wconfig.h"
#include "grouphash.h"
#include "locale.h"
#include "wdefines.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 2048
#endif

#include <fnmatch.h>

int datelen=-1;

Datei::Datei()
{
  fp=NULL;
  error=0;
}

Datei::~Datei()
{
  if(fp!=NULL) close();
}

int Datei::open(const char *name,char *mode)
{
  fp=fopen(name,mode);
  if(fp==NULL) return 1;
  return 0;
}

void Datei::close()
{
  if(fp==NULL) return;
  fclose(fp);
  fp=NULL;
}

int Datei::putUChar(unsigned char value)
{
  if(fp==NULL) return -1;
  int x;
  x=fputc(value,fp);
  if(x!=EOF) return 1;
  printf("fehler bei FPutChar\n");
  return 0;
}

int Datei::putUShort(unsigned short int value)
{
  if(fp==NULL) return -1;
  int x;
  x=fputc(value>>8,fp);
  if(x==EOF) return 0;
  x=fputc(value&0xff,fp);
  if(x==EOF) return 1;
  return 2;
}

int Datei::putInt(int value)
{
  if(fp==NULL) return -1;
  int x,tv;
  if(value<0) {
    x=fputc(1,fp);
    if(x==EOF) return 0;
  } else {
    x=fputc(0,fp);
    if(x==EOF) return 0;
  }
  tv=abs(value);
  int i;
  for(i=0;i<4;i++) {
    x=fputc(tv&0xff,fp);
    if(x==EOF) return i+1;
    tv>>=8;
  }
  return i+1;
}

int Datei::getInt()
{
  if(fp==NULL) return -1;
  int vz,x,value;
  vz=fgetc(fp);
  if(vz<0) {
    error++;
    return 0;
  }
  value=0;
  for(int i=0;i<4;i++) {
    x=fgetc(fp);
    if(x==EOF) {
      error++;
      return 0;
    }
    value+=(x<<(8*i));
  }
  if(vz==1) value*=-1;
  return value;
}

int Datei::putULong(unsigned long value)
{
  if(fp==NULL) return -1;
  int x;
  x=fputc((value&0xff000000)>>24,fp);
  if(x==EOF) return 0;
  x=fputc((value&0xff0000)>>16,fp);
  if(x==EOF) return 1;
  x=fputc((value&0xff00)>>8,fp);
  if(x==EOF) return 2;
  x=fputc(value&0xff,fp);
  if(x==EOF) return 3;
  return 4;
}

int Datei::putString(char *string)
{
  if(fp==NULL) return -1;
  int x;
  x=fputs(string,fp);
  if(x==EOF) return -2;
  return strlen(string);
}

unsigned char Datei::getUChar()
{
  if(fp==NULL) {
    error++;
    return 0;
  }
  int x;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 0;
  }
  return (unsigned char) x;
}

int Datei::getCharAsInt()
{
  if(fp==NULL) {
    error++;
    return 0;
  }
  return fgetc(fp);
}

unsigned short Datei::getUShort()
{
  if(fp==NULL) {
    error++;
    return 0;
  }
  int x;
  unsigned short int a;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 0;
  }
  a=((unsigned int)x)*0x100;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 1;
  }
  a+=(unsigned int)x;
  return a;
}

unsigned long Datei::getULong()
{
  if(fp==NULL) {
    error++;
    return 0;
  }
  int x;
  unsigned long a;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 0;
  }
  a=((unsigned int)x)*0x1000000;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 1;
  }
  a+=((unsigned int)x)*0x10000;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 2;
  }
  a+=((unsigned int)x)*0x100;
  x=fgetc(fp);
  if(x==EOF) {
    error++;
    return 3;
  }
  a+=(unsigned int)x;
  return a;
}

char *Datei::getString(int count)
{
  if(fp==NULL) return NULL;
  unsigned char *tstr,*tstr2;
  int ch;
  long x;
  long size;
  if(count==-1) {
    /* lesen bis NULL-Byte */
    size=1024;
    tstr=(unsigned char*)_allocsafe(size);
    x=0;
    do {
      ch=fgetc(fp);
      if(ch!=EOF) {
        tstr[x++]=(unsigned char)ch;
      } else {
        tstr[x++]=0;
        break;
      }
      if(x==size) {
        size*=2;
        tstr2=(unsigned char*)_allocsafe(size);
        if(tstr2==NULL) break;
        strcpy((char*)tstr2,(char*)tstr);
        _freesafe(tstr);
        tstr=tstr2;
      }
    } while(ch!=0);
  } else {
    tstr=(unsigned char*)_allocsafe(count+1);
    if(tstr!=NULL) {
      for(x=0;x<count;x++) {
        ch=fgetc(fp);
        tstr[x]=(unsigned char)ch;
      }
      tstr[count]=0;
    }
  }
  return (char*)tstr;
}

int Datei::getString(char *buf,int count)
{
  buf[0]=0;
  char *str=getString(count);
  strncpy(buf,str,count);
  buf[count]=0;
  _freesafe(str);
  return strlen(buf);
}

FileEntry::FileEntry()
{
  name=NULL;
  fullname=NULL;
  filetype=NULL;
  use=true;
  isHidden=false;
}

FileEntry::~FileEntry()
{
  if(name!=NULL) _freesafe(name);
  if(fullname!=NULL) _freesafe(fullname);
}

Verzeichnis::Verzeichnis()
{
  files=NULL;
  actdir=NULL;
  actsortmode=-1;
  lencalcs=false;
}

Verzeichnis::~Verzeichnis()
{
  closeDir();
}

ArrayList *Verzeichnis::getFiles()
{
  return files;
}

int Verzeichnis::readDir(const char *dirstr)
{
  if(dirstr==NULL) return -1;
  if(strlen(dirstr)<1) return -1;
  if(files!=NULL) closeDir();
  files=new ArrayList();
  // nahezu 1 zu 1 vom Original bernehmen

  int returnvalue=0;
  struct dirent *namelist;
  DIR *dir;
  FileEntry *fe1=NULL,*fe2;
  char *tstr;
  long x,ldnr;
  FileEntry *fpp;
  long havepointpoint;
  int id;
  bool corruptDir=false;
  
  fe1=NULL;
  dir=opendir(dirstr);
  if(dir!=NULL) {
//    readdir(dir);  /* berspringt den Eintrag "." */
    havepointpoint=-1;
    fpp=NULL;
    ldnr=0;
    while((namelist=readdir(dir))!=NULL) {
      // ignore "." entries (no need)
      if(strcmp(namelist->d_name,".")!=0) {
        fe1=new FileEntry();
        if(strcmp(namelist->d_name,"..")==0) {
          // this is because of problems with NTFS
          havepointpoint=ldnr;
          fpp=fe1;
        }
        fe1->name=dupstring(namelist->d_name);
        x=strlen(fe1->name);
        tstr=(char*)_allocsafe(strlen(dirstr)+1+x+1);
        strcpy(tstr,dirstr);
        if(strlen(dirstr)>1) strcat(tstr,"/");
        strcat(tstr,namelist->d_name);
        fe1->fullname=tstr;
        fe1->nr=ldnr++;
        if(fe1->readInfos()==0) {
          if(fe1->name[0]=='.') fe1->isHidden=true;
          else fe1->isHidden=false;
          files->addElement(fe1);
        } else {
          delete fe1;
          corruptDir=true;
        }
      }
      if(corruptDir==true) break;
    }
    closedir(dir);
    if(corruptDir==false) {
      if(havepointpoint==-1) {
        // there is no parent-dir-entry
        // create one

        // first raise the nr of all FEs
        id=files->initEnum();
        fe1=(FileEntry*)files->getFirstElement(id);
        while(fe1!=NULL) {
          fe1->nr++;
          fe1=(FileEntry*)files->getNextElement(id);
        }
        fe2=(FileEntry*)files->getFirstElement(id);
        files->closeEnum(id);
        fe1=new FileEntry();
        fe1->name=dupstring("..");
        x=strlen(fe1->name);
        tstr=(char*)_allocsafe(strlen(dirstr)+1+x+1);
        strcpy(tstr,dirstr);
        if(strlen(dirstr)>1) strcat(tstr,"/");
        strcat(tstr,fe1->name);
        fe1->fullname=tstr;
        fe1->nr=0;
        if(fe1->name[0]=='.') fe1->isHidden=true;
        else fe1->isHidden=false;
        fe1->size=0;
        fe1->dirsize=-1;
        if(fe2!=NULL) {
          fe1->lastaccess=fe2->lastaccess;
          fe1->lastmod=fe2->lastmod;
          fe1->lastchange=fe2->lastchange;
          fe1->mode=fe2->mode;
          fe1->userid=fe2->userid;
          fe1->groupid=fe2->groupid;
        }
        fe1->isCorrupt=false;
        fe1->isLink=false;
        fe1->select=false;
        files->addElementAt(0,fe1);
      } else if(havepointpoint>0) {
        if(fpp!=NULL) {
          id=files->initEnum();
          fe1=(FileEntry*)files->getFirstElement(id);
          while((fe1!=NULL)&&(fe1!=fpp)) {
            fe1->nr++;
          }
          havepointpoint=files->getIndex(fpp);
          files->removeElementAt(havepointpoint);
          files->addElementAt(0,fpp);
          fpp->nr=0;
          files->closeEnum(id);
        }
      }
      actdir=dupstring(dirstr);
    }
  } else {
    returnvalue=-1;
    delete files;
    files=NULL;
  }
  if(corruptDir==true) {
    id=files->initEnum();
    fe1=(FileEntry*)files->getFirstElement(id);
    while(fe1!=NULL) {
      delete fe1;
      fe1=(FileEntry*)files->getNextElement(id);
    }
    files->closeEnum(id);

    returnvalue=-1;
    delete files;
    files=NULL;
  }
  return returnvalue;
}

int Verzeichnis::sortfunction(void *p1,void *p2,int mode)
{
  FileEntry *fe1,*fe2;
  int comp=0;
  fe1=(FileEntry*)p1;
  fe2=(FileEntry*)p2;
  loff_t s1,s2;
  int dirm=mode&0x600;
  
/*  if(strcmp(fe1->name,"..")==0) return true;  // Sonderfall
  if(strcmp(fe2->name,"..")==0) return false;*/
  if((strcmp(fe1->name,"..")==0)&&(strcmp(fe2->name,"..")==0)) return 0;
  else if(strcmp(fe1->name,"..")==0) return -1;
  else if(strcmp(fe2->name,"..")==0) return 1;

  switch(mode&0xff) {
    case SORT_NAME:
      comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_SIZE:
      s1=-1;
      s2=-1;
      if(fe1->isDir()==true) s1=fe1->dirsize;
      if(s1<0) s1=fe1->size;
      if(fe2->isDir()==true) s2=fe2->dirsize;
      if(s2<0) s2=fe2->size;
/*      if(s1<s2) smaller=true;
      else if(s1==s2) {
        if(strcmp(fe1->name,fe2->name)<0) smaller=true;
      }*/
      if(s1<s2) comp=-1;
      else if(s1>s2) comp=1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_ACCTIME:
      if(fe1->lastaccess<fe2->lastaccess) comp=-1;
      else if(fe1->lastaccess>fe2->lastaccess) comp=1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_MODTIME:
      if(fe1->lastmod<fe2->lastmod) comp=-1;
      else if(fe1->lastmod>fe2->lastmod) comp=1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_CHGTIME:
      if(fe1->lastchange<fe2->lastchange) comp=-1;
      else if(fe1->lastchange>fe2->lastchange) comp=1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
  }
  if((mode&SORT_REVERSE)==SORT_REVERSE) comp*=-1;
  if(dirm!=SORT_DIRMIXED) {
    if(fe1->isDir()==true) {
      if(fe2->isDir()==false) comp=(dirm==SORT_DIRLAST)?1:-1;
    } else {
      if(fe2->isDir()==true) comp=(dirm==SORT_DIRLAST)?-1:1;
    }
  }
  return comp;
}

void Verzeichnis::closeDir()
{
  if(files!=NULL) {
    FileEntry *fe=(FileEntry*)files->getFirstElement();
    while(fe!=NULL) {
      delete fe;
      fe=(FileEntry*)files->getNextElement();
    }
    delete files;
    files=NULL;
    _freesafe(actdir);
    actdir=NULL;
    actsortmode=-1;
    lencalcs=false;
  }
}

int Verzeichnis::getSize()
{
  if(files!=NULL) return files->size();
  else return -1;
}

char *Verzeichnis::getDir()
{
  return actdir;
}

char *ParentDir(const char *pathstr,char *dirnamereturn)
{
  char *newpath;
  int x,y;
  x=strlen(pathstr);
  if(x<2) {
    newpath=(char*)_allocsafe(2);
    strcpy(newpath,"/");
    if(dirnamereturn!=NULL) strcpy(dirnamereturn,"");
  } else {
    if(pathstr[x-1]=='/') y=x-2; else y=x-1;
    while((pathstr[y]!='/')&&(y>0)) {
      y--;
    }
    if(y==0) {
      newpath=(char*)_allocsafe(2);
      strcpy(newpath,"/");
      if(dirnamereturn!=NULL) strcpy(dirnamereturn,pathstr+1);
    } else {
      newpath=(char*)_allocsafe(y+1);
      strncpy(newpath,pathstr,y);
      if(dirnamereturn!=NULL) strcpy(dirnamereturn,pathstr+y+1);
      newpath[y]=0;
    }
  }
  return newpath;
}

char *HandlePath(const char *pathstr)
{
  char *newpath,*newpath2,buffer[1024];
  int x,y,parentlevel;
  int start,end;
  int found,ende;
  x=strlen(pathstr);
  y=0;
  found=0;
  parentlevel=0;
  ende=0;
  newpath=(char*)_allocsafe(1);
  newpath[0]=0;
  do {
    start=y;
    while((pathstr[y]!='/')&&((y+1)<x)) {
      y++;
    }
    if(((y+1)>=x)&&(pathstr[y]!='/')) end=y+1; else end=y;
    if(end>start) {
      strncpy(buffer,pathstr+start,end-start);
      buffer[end-start]=0;
      if(strcmp(buffer,".")==0) {
        y++;
      } else if(strcmp(buffer,"..")==0) {
        y++;
        newpath2=ParentDir(newpath,NULL);
        _freesafe(newpath);
        newpath=newpath2;
      } else {
        newpath2=(char*)_allocsafe(strlen(newpath)+1+strlen(buffer)+1);
        strcpy(newpath2,newpath);
        if(strlen(newpath)>0) {
          if(newpath[strlen(newpath)-1]!='/') strcat(newpath2,"/");
        }
        strcat(newpath2,buffer);
        _freesafe(newpath);
        newpath=newpath2;
        y++;
      }
    } else {
      if(start==0) {
        newpath2=(char*)_allocsafe(2);
        strcpy(newpath2,"/");
        _freesafe(newpath);
        newpath=newpath2;
      }
      y++;
    }
    if(y>=x) ende=1;
  } while(ende==0);
  return newpath;
}

int Verzeichnis::sort(int mode)
{
  return sort(mode,false);
}

int Verzeichnis::sort(int mode,bool force)
{
 if((mode!=actsortmode)||(force==true)) {
   files->sort(sortfunction,mode);
   // FEs erhalten Nummer anhand der Position, damit von LVC auf FE geschlossen werden kann
   int id=files->initEnum(),pos;
   FileEntry *fe=(FileEntry*)files->getFirstElement(id);
   pos=0;
   while(fe!=NULL) {
     fe->nr=pos++;
     fe=(FileEntry*)files->getNextElement(id);
   }
   files->closeEnum(id);
   actsortmode=mode;
 }
 return 0;
}

bool FileEntry::isDir()
{
  bool dir=false;
  if(isLink==true) {
    if(S_ISDIR(dmode)) dir=true;
  } else if(S_ISDIR(mode)) dir=true;
  return dir;
}

FileEntry *FileEntry::duplicate()
{
  FileEntry *tfe=new FileEntry();
  tfe->fullname=dupstring(fullname);
  tfe->name=dupstring(name);
  tfe->size=size;
  tfe->dsize=dsize;
  tfe->dirsize=dirsize;
  tfe->select=select;
  tfe->lastaccess=lastaccess;
  tfe->dlastaccess=dlastaccess;
  tfe->lastmod=lastmod;
  tfe->dlastmod=dlastmod;
  tfe->lastchange=lastchange;
  tfe->dlastchange=dlastchange;
  tfe->mode=mode;
  tfe->dmode=dmode;
  tfe->isLink=isLink;
  tfe->isCorrupt=isCorrupt;
  tfe->userid=userid;
  tfe->duserid=duserid;
  tfe->groupid=groupid;
  tfe->dgroupid=dgroupid;
  tfe->nr=nr;
  tfe->inFilter=inFilter;
  tfe->filetype=filetype;
  return tfe;
}

bool Datei::fileExists(const char *name)
{
  struct stat stbuf;
  if(lstat(name,&stbuf)!=0) return false;
  else {
    /* kann auch verzeichnis sein */
//    if(S_ISDIR(stbuf.st_mode)) return false;
  }
  return true;
}

Datei::d_fe_t Datei::fileExistsExt(const char*name)
{
  struct stat stbuf;
  d_fe_t return_value;
  if(stat(name,&stbuf)!=0) return_value=D_FE_NOFILE;
  else {
    if(S_ISDIR(stbuf.st_mode)) return_value=D_FE_DIR;
    else return_value=D_FE_FILE;
    if(S_ISLNK(stbuf.st_mode)) return_value=D_FE_LINK;
  }
  return return_value;
}

Datei::d_fe_t Datei::lfileExistsExt(const char*name)
{
  struct stat stbuf;
  d_fe_t return_value;
  if(lstat(name,&stbuf)!=0) return_value=D_FE_NOFILE;
  else {
    if(S_ISDIR(stbuf.st_mode)) return_value=D_FE_DIR;
    else return_value=D_FE_FILE;
    if(S_ISLNK(stbuf.st_mode)) return_value=D_FE_LINK;
  }
  return return_value;
}

void Datei::seek(long offset,int mode)
{
  if(fp!=NULL) fseek(fp,offset,mode);
}

void Verzeichnis::recalcLengths()
{
  int maxnamelen=-1,
      maxsizelen=-1,
      maxtypelen=-1,
      maxownerlen=-1,
      maxdestlen=-1;
  #ifdef DEBUG
  printf("entering Verzeichnis::recalcLengths()\n");
  #endif
  int id=files->initEnum(),tlen;
  bool usefiletypelen=true;
  char *str;
  loff_t size;
  
  FileEntry *fe=(FileEntry*)files->getFirstElement(id);
  while(fe!=NULL) {
    tlen=strlen(fe->name);
    if(tlen>maxnamelen) maxnamelen=tlen;

    if((fe->isDir()==true)&&(fe->isLink==false)&&(fe->dirsize>=0)) size=fe->dirsize;
    else size=fe->size;
    tlen=LongSize(size);
    if(tlen>maxsizelen) maxsizelen=tlen;

  #ifdef DEBUG
//  printf("  getOwnerStringLen\n");
  #endif
    tlen=getOwnerStringLen(fe->userid,fe->groupid);
  #ifdef DEBUG
//  printf("  getOwnerStringLen ok\n");
  #endif
    if(tlen>maxownerlen) maxownerlen=tlen;
    if((fe->filetype!=NULL)&&(usefiletypelen==true)) {
//      tlen=strlen(fe->filetype->getName());
      tlen=fe->filetype->getNameLen();
      if(tlen>maxtypelen) maxtypelen=tlen;
    } else {
      // Ein Typ wurde noch nicht erkannt => Nicht weiterchecken, sondern Maximalwert benutzen
      usefiletypelen=false;
    }
    str=fe->getDestination();
    if(str!=NULL) {
      tlen=strlen(str);
      if(tlen>maxdestlen) maxdestlen=tlen;
      _freesafe(str);
    }
    fe=(FileEntry*)files->getNextElement(id);
  }
  files->closeEnum(id);
  tlen=strlen(catalog.getLocale(109)); // Dir-Name
  if(tlen>maxtypelen) maxtypelen=tlen;
  namelen=maxnamelen;
  sizelen=maxsizelen;
  if(usefiletypelen==true) {
    typelen=maxtypelen;
  } else {
    typelen=wconfig->getMaxFiletypeLen();
  }
  ownerlen=maxownerlen;
  destlen=maxdestlen;
  lencalcs=true;
  #ifdef DEBUG
  printf("leaving Verzeichnis::recalcLengths\n");
  #endif
}

void Verzeichnis::recalcLengths(int *sets)
{
  int maxnamelen=-1,showname=0,
      maxsizelen=-1,showsize=0,
      maxtypelen=-1,showtype=0,
      maxownerlen=-1,showowner=0,
      maxdestlen=-1,showdest=0;
  int id=files->initEnum(),tlen;
  char *str;
  bool usefiletypelen=true;
  loff_t size;

  for(int i=0;i<9;i++) {
    if(sets[i]<0) break;
    switch(sets[i]) {
      case 0://Name
        showname=1;
        break;
      case 1://Size
        showsize=1;
        break;
      case 2://Type
        showtype=1;
        break;
      case 4://Owner
        showowner=1;
        break;
      case 5://Dest
        showdest=1;
        break;
    }
  }

  FileEntry *fe=(FileEntry*)files->getFirstElement(id);
  while(fe!=NULL) {
    if(showname==1) {
      tlen=strlen(fe->name);
      if(tlen>maxnamelen) maxnamelen=tlen;
    }
    if(showsize==1) {
      if((fe->isDir()==true)&&(fe->isLink==false)&&(fe->dirsize>=0)) size=fe->dirsize;
      else size=fe->size;
      tlen=LongSize(size);
      if(tlen>maxsizelen) maxsizelen=tlen;
    }
    if(showowner==1) {
      tlen=getOwnerStringLen(fe->userid,fe->groupid);
      if(tlen>maxownerlen) maxownerlen=tlen;
    }
    if((showtype==1)&&(fe->filetype!=NULL)&&(usefiletypelen==true)) {
//      tlen=strlen(fe->filetype->getName());
      tlen=fe->filetype->getNameLen();
      if(tlen>maxtypelen) maxtypelen=tlen;
    } else {
      // Ein Typ wurde noch nicht erkannt => Nicht weiterchecken, sondern Maximalwert benutzen
      usefiletypelen=false;
    }
    if(showdest==1) {
      str=fe->getDestination();
      if(str!=NULL) {
        tlen=strlen(str);
        if(tlen>maxdestlen) maxdestlen=tlen;
        _freesafe(str);
      }
    }
    fe=(FileEntry*)files->getNextElement(id);
  }
  files->closeEnum(id);
  namelen=maxnamelen;
  sizelen=maxsizelen;
  tlen=strlen(catalog.getLocale(109)); // Dir-Name
  if(tlen>maxtypelen) maxtypelen=tlen;
  if(usefiletypelen==true) {
    typelen=maxtypelen;
  } else {
    typelen=wconfig->getMaxFiletypeLen();
  }
  ownerlen=maxownerlen;
  destlen=maxdestlen;
  lencalcs=true;
}

int getOwnerStringLen(uid_t user,gid_t gr)
{
  char *tstr;
  #ifdef DEBUG
//  printf("entering getOwnerStringLen\n");
  #endif
  int len=0;
  int ost=wconfig->getOwnerstringtype();
  #ifdef DEBUG
//  printf("  getUserNameS\n");
  #endif
  tstr=ugdb->getUserNameS(user);
  if(tstr!=NULL) len+=strlen(tstr);
  len+=(ost==1)?1:3;
  #ifdef DEBUG
//  printf("  getGroupNameS\n");
  #endif
  tstr=ugdb->getGroupNameS(gr);
  if(tstr!=NULL) len+=strlen(tstr);
  #ifdef DEBUG
//  printf("leaving getOwnerStringLen\n");
  #endif
  return len;
}

char* getOwnerString(uid_t user,gid_t mygroup)
{
  int len=0;
  char *tstr1,*tstr2;
  int ost=wconfig->getOwnerstringtype();
  
  tstr1=ugdb->getUserNameS(user);
  tstr2=ugdb->getGroupNameS(mygroup);
  if(tstr1!=NULL) len+=strlen(tstr1);
  len+=(ost==1)?1:3;
  if(tstr2!=NULL) len+=strlen(tstr2);
  char *str=(char*)_allocsafe((len+2)*sizeof(char));
  str[0]=0;
  if(tstr1!=NULL) strcat(str,tstr1);
  strcat(str,(ost==1)?".":" @ ");
  if(tstr2!=NULL) strcat(str,tstr2);
  return str;
}

int FileEntry::readInfos()
{
  return readInfos(false);
}

int FileEntry::readInfos(bool update)
{
  //TODO: auf update achten
  struct stat stbuf;

  select=false;
  dirsize=-1;
  filetype=NULL;
  inFilter=true;

  if(fullname==NULL) return 1;
  if(lstat(fullname,&stbuf)==0) {
    size=stbuf.st_size;
    lastaccess=stbuf.st_atime;
    lastmod=stbuf.st_mtime;
    lastchange=stbuf.st_ctime;
    mode=stbuf.st_mode;
    userid=stbuf.st_uid;
    groupid=stbuf.st_gid;
  } else return 2;
  isCorrupt=false;
  if(S_ISLNK(mode)) {
    isLink=true;
    if(stat(fullname,&stbuf)==0) {
      dsize=stbuf.st_size;
      dlastaccess=stbuf.st_atime;
      dlastmod=stbuf.st_mtime;
      dlastchange=stbuf.st_ctime;
      dmode=stbuf.st_mode;
      duserid=stbuf.st_uid;
      dgroupid=stbuf.st_gid;
    } else {
      /* korrupt */
      isCorrupt=true;
      if(errno==ENOENT) {
        // Eventl. fuer spaetere Betrachtungen
      }
    }
  } else {
    isLink=false;
  }
  return 0;
}

char *FileEntry::getDestination()
{
  char *str=NULL,*tstr;
  if((isLink==true)&&(fullname!=NULL)) {
    tstr=(char*)_allocsafe((MAXPATHLEN+1)*sizeof(char));
    int x=readlink(fullname,tstr,MAXPATHLEN);
    if(x>=0) {
      tstr[x]=0;
      str=dupstring(tstr);
    }
    _freesafe(tstr);
  }
  return str;
}

int FileEntry::writeDestination2Buf(char *buf,int size2,bool terminated)
{
  int x=-1;
  if((isLink==true)&&(fullname!=NULL)) {
    x=readlink(fullname,buf,size2);
    if((terminated==true)&&(x>=0)) {
      buf[x]=0;
    }
  }
  return x;
}

ListViewContent *Verzeichnis::getLVC4FE(FileEntry *fe,int *sets)
{
  int len=0,mlen=0,x,dl;
  char *str,*tstr,buf[128],*tstr2,buf2[128],*tstr3;
  struct tm *timeptr;
//struct timeval tv1,tv2;
  dl=getDatelen();
  //TODO
  printf("Warning: ListViewContent::getLVC4FE not complete implemented yet\n");
  if((lencalcs==true)&&(1==0)) {
    // schneller, es muessen aber alle Werte berechnet sein
    if(namelen>=0) mlen+=namelen+1;
    if(sizelen>=0) mlen+=sizelen+1;
    if(typelen>=0) mlen+=typelen+1;
    mlen+=10;
    if(ownerlen>=0) mlen+=ownerlen+1;
    if(destlen>=0) mlen+=destlen+1;
    mlen+=3*(dl+1);
    mlen+=1;
    mlen+=10; // Sicherheitszone
    str=(char*)_allocsafe(mlen*sizeof(char));
    str[0]=0;
    for(int i=0;i<9;i++) {
      if(sets[i]<0) break;
      switch(sets[i]) {
        case 0://Name
//gettimeofday(&tv1,NULL);
          sprintf(buf2,"%%s%%s%%-%ds ",namelen);
          buf[0]=' ';
          buf[1]=0;
          if(fe->isLink==true) {
            if(fe->isCorrupt==true) buf[0]='!';
            else {
              if(S_ISDIR(fe->dmode)) buf[0]='~';
              else buf[0]='@';
            }
          } else {
            if(S_ISDIR(fe->mode)) buf[0]='/';
            else if(fe->isExe()==true) buf[0]='*';
          }
          sprintf(str,buf2,str,buf,fe->name);
//gettimeofday(&tv2,NULL);
//printf("%d %d\n",tv1.tv_sec-tv2.tv_sec,tv1.tv_usec-tv2.tv_usec);
          break;
        case 1://Size
          sprintf(buf2,"%%s%%-%ds ",sizelen);
          buf[MakeLong2NiceStr(buf,fe->size)]='\0';
          sprintf(str,buf2,str,buf);
          break;
        case 2://Type
          if(fe->filetype!=NULL) {
            sprintf(buf2,"%%s%%-%ds ",typelen);
            sprintf(str,buf2,str,fe->filetype->getName());
          } else {
            sprintf(buf2,"%%s%%-%ds ",typelen);
            sprintf(str,buf2,str,"");
          }
          break;
        case 3://Perm
          x=strlen(str);
          fe->writePerm2Buf(str+x,true);
          str[x+9]=' ';
          str[x+1+9]=0;
          break;
        case 4://Owner
          tstr2=getOwnerString(fe->userid,fe->groupid);
          sprintf(buf2,"%%s%%-%ds ",ownerlen);
          sprintf(str,buf2,str,tstr2);
          _freesafe(tstr2);
          break;
/*        case 5://Dest
          len=destlen;
          tstr3=fe->getDestination();
          if((len<0)&&(tstr3!=NULL)) {
            len=strlen(tstr3);
          }
          if(len>=0) {
            tstr=(char*)_allocsafe((len+1)*sizeof(char));
            sprintf(buf,"%%-%ds",len);
            if(tstr3!=NULL) sprintf(tstr,buf,tstr3);
            else sprintf(tstr,buf,"");
            tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
            if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
            else sprintf(tstr2,"%s",tstr);
            _freesafe(str);
            _freesafe(tstr);
            str=tstr2;
          }
          if(tstr3!=NULL) _freesafe(tstr3);
          break;*/
        case 6://Mod
          timeptr=localtime(&(fe->lastmod));
          tstr2=asctime(timeptr);
          tstr2[strlen(tstr2)-1]=0;
          sprintf(buf2,"%%s%%-%ds ",dl);
          sprintf(str,buf2,str,tstr2);
          break;
        case 7://Acc
          timeptr=localtime(&(fe->lastaccess));
          tstr2=asctime(timeptr);
          tstr2[strlen(tstr2)-1]=0;
          sprintf(buf2,"%%s%%-%ds ",dl);
          sprintf(str,buf2,str,tstr2);
          break;
        case 8://Change
          timeptr=localtime(&(fe->lastchange));
          tstr2=asctime(timeptr);
          tstr2[strlen(tstr2)-1]=0;
          sprintf(buf2,"%%s%%-%ds ",dl);
          sprintf(str,buf2,str,tstr2);
          break;
      }
    }
  } else {
    str=dupstring("");
    for(int i=0;i<9;i++) {
      if(sets[i]<0) break;
      switch(sets[i]) {
        case 0://Name
          len=namelen;
          if(len<0) len=strlen(fe->name);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,fe->name);
          buf[0]=' ';
          buf[1]=0;
          if(fe->isLink==true) {
            if(fe->isCorrupt==true) buf[0]='!';
            else {
              if(S_ISDIR(fe->dmode)) buf[0]='~';
              else buf[0]='@';
            }
          } else {
            if(S_ISDIR(fe->mode)) buf[0]='/';
            else if(fe->isExe()==true) buf[0]='*';
          }
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(buf)+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s%s",str,buf,tstr);
          else sprintf(tstr2,"%s%s",buf,tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 1://Size
          len=sizelen;
          if(len<0) len=LongSize(fe->size);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          buf2[MakeLong2NiceStr(buf2,fe->size)]='\0';
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,buf2);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 2://Type
          len=typelen;
          if((len<0)&&(fe->filetype!=NULL)) len=strlen(fe->filetype->getName());
          if(len>=0) {
            tstr=(char*)_allocsafe((len+1)*sizeof(char));
            sprintf(buf,"%%-%ds",len);
            sprintf(tstr,buf,fe->filetype->getName());
            tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
            if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
            else sprintf(tstr2,"%s",tstr);
            _freesafe(str);
            _freesafe(tstr);
            str=tstr2;
          }
          break;
        case 3://Perm
          tstr=fe->getPermStr();
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 4://Owner
          len=ownerlen;
          if(len<0) len=getOwnerStringLen(fe->userid,fe->groupid);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          tstr2=getOwnerString(fe->userid,fe->groupid);
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr2);
          _freesafe(tstr2);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 5://Dest
          len=destlen;
          tstr3=fe->getDestination();
          if((len<0)&&(tstr3!=NULL)) {
            len=strlen(tstr3);
          }
          if(len>=0) {
            tstr=(char*)_allocsafe((len+1)*sizeof(char));
            sprintf(buf,"%%-%ds",len);
            if(tstr3!=NULL) sprintf(tstr,buf,tstr3);
            else sprintf(tstr,buf,"");
            tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
            if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
            else sprintf(tstr2,"%s",tstr);
            _freesafe(str);
            _freesafe(tstr);
            str=tstr2;
          }
          if(tstr3!=NULL) _freesafe(tstr3);
          break;
        case 6://Mod
          timeptr=localtime(&(fe->lastmod));
          tstr3=asctime(timeptr);
          len=dl;
          if(len<0) len=strlen(tstr3);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr3);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          str[strlen(str)-1]=0;
          break;
        case 7://Acc
          timeptr=localtime(&(fe->lastaccess));
          tstr3=asctime(timeptr);
          len=dl;
          if(len<0) len=strlen(tstr3);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr3);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          str[strlen(str)-1]=0;
          break;
        case 8://Change
          timeptr=localtime(&(fe->lastchange));
          tstr3=asctime(timeptr);
          len=dl;
          if(len<0) len=strlen(tstr3);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr3);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          str[strlen(str)-1]=0;
          break;
      }
    }
  }
  ListViewContent *lvc=new ListViewContent(str,1,1,1,1,0,0,0,0,0);
  _freesafe(str);
  return lvc;
}

extern double d1,d2,d3;

void Verzeichnis::setLVC4FE(ListViewContent *lvc,FileEntry *fe,int *sets)
{
  int len=0,mlen=0,x,x2,pos,dl;
  char *tstr,buf[128],*tstr2,buf2[128],*tstr3;
  char *str=NULL;
  struct tm *timeptr;
  loff_t size;
  int newlengths[5],ost;
  
  dl=getDatelen();
  if(lencalcs==true) {
    // schneller, es muessen aber alle Werte berechnet sein
    // this calc all need lengths of the fe so we doesn't need to calc it later
    // this will also sets the *len variables correctly if a entry growed
    calcLength(fe,sets,newlengths);
    if(namelen>=0) mlen+=namelen+1;
    if(sizelen>=0) mlen+=sizelen+1;
    if(typelen>=0) mlen+=typelen+1;
    mlen+=10;
    if(ownerlen>=0) mlen+=ownerlen+1;
    if(destlen>=0) mlen+=destlen+1;
    mlen+=3*(dl+1);
    mlen+=1;
    mlen+=10; // Sicherheitszone
#ifdef DEBUG
//    printf("use len in setLVC4FE: %d\n",mlen);
#endif
    str=(char*)_allocsafe(mlen*sizeof(char));
    memset(str,' ',mlen-1);
    pos=0;
    for(int i=0;i<9;i++) {
      if(sets[i]<0) break;
      switch(sets[i]) {
        case 0://Name
          if(fe->isLink==true) {
            if(fe->isCorrupt==true) str[pos]='!';
            else {
              if(S_ISDIR(fe->dmode)) str[pos]='~';
              else str[pos]='@';
            }
          } else {
            if(S_ISDIR(fe->mode)) str[pos]='/';
            else if(fe->isExe()==true) str[pos]='*';
          }
//          x=strlen(fe->name);
          x=newlengths[V_NAMELENGTH];
          memcpy(str+pos+1,fe->name,x);
          pos+=namelen+2;
          break;
        case 1://Size
          if((fe->isDir()==true)&&(fe->isLink==false)&&(fe->dirsize>=0)) size=fe->dirsize;
          else size=fe->size;
#ifdef LEFTJUSTIFY
          MakeLong2NiceStr(str+pos,size);
          str[pos+newlengths[V_SIZELENGTH]]=' ';
#else
          MakeLong2NiceStr(str+pos+sizelen-newlengths[V_SIZELENGTH],size);
          str[pos+sizelen]=' ';
#endif
//          str[pos+LongSize(size)]=' ';
          pos+=sizelen+1;
          break;
        case 2://Type
          if(fe->filetype!=NULL) {
//            x=strlen(fe->filetype->getName());
//            x=fe->filetype->getNameLen();
            x=newlengths[V_TYPELENGTH];
            memcpy(str+pos,fe->filetype->getName(),x);
          } else {
            if(fe->isDir()==true) {
              // Dir
              //TODO:nicht jedesmal strlen berechnen muessen
//              x=strlen(catalog.getLocale(109));
//              memcpy(str+pos,catalog.getLocale(109),x);
              x=wconfig->getdirtype()->getNameLen();
              memcpy(str+pos,wconfig->getdirtype()->getName(),x);
            } else {
//              x=strlen(wconfig->getnotyettype()->getName());
              x=wconfig->getnotyettype()->getNameLen();
              memcpy(str+pos,wconfig->getnotyettype()->getName(),x);
            }
          }
          pos+=typelen+1;
          break;
        case 3://Perm
          fe->writePerm2Buf(str+pos,false);
          pos+=10;
          break;
        case 4://Owner
          // will be 1 for Owner.group, 0 else
          ost=wconfig->getOwnerstringtype();
          tstr=ugdb->getUserNameS(fe->userid);
          x2=pos;
          if(tstr!=NULL) {
            x=strlen(tstr);
            memcpy(str+x2,tstr,x);
            x2+=x;
          }
          if(ost==1) {
            str[x2++]='.';
          } else {
            x2++;
            str[x2++]='@';
            x2++;
          }
          tstr=ugdb->getGroupNameS(fe->groupid);
          if(tstr!=NULL) {
            x=strlen(tstr);
            memcpy(str+x2,tstr,x);
          }
          pos+=ownerlen+1;
          break;
        case 5://Dest
          fe->writeDestination2Buf(str+pos,destlen,false);
          pos+=destlen+1;
          break;
        case 6://Mod
          timeptr=localtime(&(fe->lastmod));
          x=strftime(str+pos,dl+1,"%H:%M:%S %d %b %Y",timeptr);
          str[pos+x]=' ';
          pos+=dl+1;
          break;
        case 7://Acc
          timeptr=localtime(&(fe->lastaccess));
          x=strftime(str+pos,dl+1,"%H:%M:%S %d %b %Y",timeptr);
          str[pos+x]=' ';
          pos+=dl+1;
          break;
        case 8://Change
          timeptr=localtime(&(fe->lastchange));
          x=strftime(str+pos,dl+1,"%H:%M:%S %d %b %Y",timeptr);
          str[pos+x]=' ';
          pos+=dl+1;
          break;
      }
    }
    str[((pos>0)?(pos-1):pos)]=0;
  } else {
    printf("Warning:Wrong else in setLVC\nPlease notice the author about this!\n");
    str=dupstring("");
    for(int i=0;i<9;i++) {
      if(sets[i]<0) break;
      switch(sets[i]) {
        case 0://Name
          len=namelen;
          if(len<0) len=strlen(fe->name);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,fe->name);
          buf[0]=' ';
          buf[1]=0;
          if(fe->isLink==true) {
            if(fe->isCorrupt==true) buf[0]='!';
            else {
              if(S_ISDIR(fe->dmode)) buf[0]='~';
              else buf[0]='@';
            }
          } else {
            if(S_ISDIR(fe->mode)) buf[0]='/';
            else if(fe->isExe()==true) buf[0]='*';
          }
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(buf)+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s%s",str,buf,tstr);
          else sprintf(tstr2,"%s%s",buf,tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 1://Size
          len=sizelen;
          if(len<0) len=LongSize(fe->size);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          buf2[MakeLong2NiceStr(buf2,fe->size)]='\0';
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,buf2);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 2://Type
          len=typelen;
          //TODO: Hier muss auch auf notyet geachtet werden, im Moment zwar nicht so
          //      schlimm da dies nicht benutzt wird (weil zu LANGSAM) aber trotzdem
          printf("Warning:Wrong case in setLVC\nPlease notice the author about this!\n");
          if((len<0)&&(fe->filetype!=NULL)) len=strlen(fe->filetype->getName());
          if(len>=0) {
            tstr=(char*)_allocsafe((len+1)*sizeof(char));
            sprintf(buf,"%%-%ds",len);
            sprintf(tstr,buf,fe->filetype->getName());
            tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
            if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
            else sprintf(tstr2,"%s",tstr);
            _freesafe(str);
            _freesafe(tstr);
            str=tstr2;
          }
          break;
        case 3://Perm
          tstr=fe->getPermStr();
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 4://Owner
          len=ownerlen;
          if(len<0) len=getOwnerStringLen(fe->userid,fe->groupid);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          tstr2=getOwnerString(fe->userid,fe->groupid);
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr2);
          _freesafe(tstr2);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          break;
        case 5://Dest
          len=destlen;
          tstr3=fe->getDestination();
          if((len<0)&&(tstr3!=NULL)) {
            len=strlen(tstr3);
          }
          if(len>=0) {
            tstr=(char*)_allocsafe((len+1)*sizeof(char));
            sprintf(buf,"%%-%ds",len);
            if(tstr3!=NULL) sprintf(tstr,buf,tstr3);
            else sprintf(tstr,buf,"");
            tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
            if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
            else sprintf(tstr2,"%s",tstr);
            _freesafe(str);
            _freesafe(tstr);
            str=tstr2;
          }
          if(tstr3!=NULL) _freesafe(tstr3);
          break;
        case 6://Mod
          timeptr=localtime(&(fe->lastmod));
          tstr3=asctime(timeptr);
          len=dl;
          if(len<0) len=strlen(tstr3);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr3);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          str[strlen(str)-1]=0;
          break;
        case 7://Acc
          timeptr=localtime(&(fe->lastaccess));
          tstr3=asctime(timeptr);
          len=dl;
          if(len<0) len=strlen(tstr3);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr3);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          str[strlen(str)-1]=0;
          break;
        case 8://Change
          timeptr=localtime(&(fe->lastchange));
          tstr3=asctime(timeptr);
          len=dl;
          if(len<0) len=strlen(tstr3);
          tstr=(char*)_allocsafe((len+1)*sizeof(char));
          sprintf(buf,"%%-%ds",len);
          sprintf(tstr,buf,tstr3);
          tstr2=(char*)_allocsafe((strlen(str)+1+strlen(tstr)+2)*sizeof(char));
          if(strlen(str)>0) sprintf(tstr2,"%s %s",str,tstr);
          else sprintf(tstr2,"%s",tstr);
          _freesafe(str);
          _freesafe(tstr);
          str=tstr2;
          str[strlen(str)-1]=0;
          break;
      }
    }
  }
  lvc->useTextQ(str);
}

void Verzeichnis::calcLength(FileEntry *fe,int *sets,int return_lengths [])
{
  int maxnamelen=-1,showname=0,
      maxsizelen=-1,showsize=0,
      maxtypelen=-1,showtype=0,
      maxownerlen=-1,showowner=0,
      maxdestlen=-1,showdest=0;
  char *str;
  loff_t size;

  for(int i=0;i<9;i++) {
    if(sets[i]<0) break;
    switch(sets[i]) {
      case 0://Name
        showname=1;
        break;
      case 1://Size
        showsize=1;
        break;
      case 2://Type
        showtype=1;
        break;
      case 4://Owner
        showowner=1;
        break;
      case 5://Dest
        showdest=1;
        break;
    }
  }

  if(showname==1) maxnamelen=strlen(fe->name);
  if(showsize==1) {
    if((fe->isDir()==true)&&(fe->isLink==false)&&(fe->dirsize>=0)) size=fe->dirsize;
    else size=fe->size;
    maxsizelen=LongSize(size);
  }
  if(showowner==1) maxownerlen=getOwnerStringLen(fe->userid,fe->groupid);
  if((showtype==1)&&(fe->filetype!=NULL)) {
    maxtypelen=fe->filetype->getNameLen();
  } else {
    maxtypelen=wconfig->getMaxFiletypeLen();
  }
  if(showdest==1) {
    str=fe->getDestination();
    if(str!=NULL) {
      maxdestlen=strlen(str);
      _freesafe(str);
    }
  }

  // store values in return_lengths if not NULL
  if(return_lengths!=NULL) {
    return_lengths[V_NAMELENGTH]=maxnamelen;
    return_lengths[V_SIZELENGTH]=maxsizelen;
    return_lengths[V_TYPELENGTH]=maxtypelen;
    return_lengths[V_OWNERLENGTH]=maxownerlen;
    return_lengths[V_DESTLENGTH]=maxdestlen;
  }
  
  //now update global values when lencalcs is true
  if(lencalcs==true) {
    if(maxnamelen>namelen) namelen=maxnamelen;
    if(maxsizelen>sizelen) sizelen=maxsizelen;
    if(maxtypelen>typelen) typelen=maxtypelen;
    if(maxownerlen>ownerlen) ownerlen=maxownerlen;
    if(maxdestlen>destlen) destlen=maxdestlen;
  }
}

char *FileEntry::getPermStr()
{
  return dupstring(getSPermStr());
}

char *FileEntry::getSPermStr()
{
  static char str[10];
  writePerm2Buf(str,true);
  return str;
}

void FileEntry::writePerm2Buf(char*str,bool terminated)
{
  if((mode&S_IRUSR)==S_IRUSR) str[0]='r'; else str[0]='-';
  if((mode&S_IWUSR)==S_IWUSR) str[1]='w'; else str[1]='-';
  if((mode&S_ISUID)==S_ISUID) {
    if((mode&S_IXUSR)==S_IXUSR) str[2]='s'; else str[2]='S';
  } else {
    if((mode&S_IXUSR)==S_IXUSR) str[2]='x'; else str[2]='-';
  }
  if((mode&S_IRGRP)==S_IRGRP) str[3]='r'; else str[3]='-';
  if((mode&S_IWGRP)==S_IWGRP) str[4]='w'; else str[4]='-';
  if((mode&S_ISGID)==S_ISGID) {
    if((mode&S_IXGRP)==S_IXGRP) str[5]='s'; else str[5]='S';
  } else {
    if((mode&S_IXGRP)==S_IXGRP) str[5]='x'; else str[5]='-';
  }
  if((mode&S_IROTH)==S_IROTH) str[6]='r'; else str[6]='-';
  if((mode&S_IWOTH)==S_IWOTH) str[7]='w'; else str[7]='-';
  if((mode&S_ISVTX)==S_ISVTX) {
    if((mode&S_IXOTH)==S_IXOTH) str[8]='t'; else str[8]='T';
  } else {
    if((mode&S_IXOTH)==S_IXOTH) str[8]='x'; else str[8]='-';
  }
  if(terminated==true) str[9]=0;
}

bool FileEntry::isExe()
{
  if((mode&S_IXUSR)==S_IXUSR) return true;
  if((mode&S_IXGRP)==S_IXGRP) return true;
  if((mode&S_IXOTH)==S_IXOTH) return true;
  return false;
}

int getDatelen()
{
  char str[128];
  if(datelen<0) {
    /* Erst berechnen */
    time_t t1;
    struct tm *mytm;
    int x;
    time(&t1);
    // Alle 7 Tage messen
    for(int i=0;i<7;i++) {
      mytm=localtime(&t1);
      x=strftime(str,127,"%H:%M:%S %d %b %Y",mytm);
      if(x>datelen) datelen=x;
      t1+=60*60*24;  // Ein Tag drauf
    }
  }
  return datelen;
}

WCFiletype* FileEntry::checkFiletype()
{
//printf("start check %s\n",name);
  unsigned char fbuffer[64];
  int reads,x,entry,found,y,minsize=0;
  int id;
  WCFiletype *ft,*unknownft;
#ifdef HAVE_LIBZ
  gzFile fp;
#else
  FILE *fp;
#endif
  List *ftlist=wconfig->getFiletypes();
  // TODO: Nicht superideal, weil position nicht zwingend richtig ist (im Moment
  //        klappt das, aber wer weiss, was mir noch alles fuer dummkramm einfaellt :-)
  //       Besser nach UNKNOWNTYPE suchen, aber zeitintensiv
  //       Viel besser: bei initList die Position garantieren
  id=ftlist->initEnum();
  unknownft=(WCFiletype*)ftlist->getElementAt(id,UNKNOWNTYPE-1);
  ftlist->closeEnum(id);
  // Vorbesetzen, damit immer unknown typ kommt, selbst wenn Datei nicht den Anforderungen
  //  fuer Dateityperkennung genuegt
  filetype=unknownft;
  if(ftlist->size()<4) {
    return filetype;
  }
  if((S_ISDIR(mode))&&(isCorrupt==false)) {
    // Wirklich ein Verzeichnis
    return filetype;
  }
  if( ((isLink==true)&&(!S_ISREG(dmode))) || ((isLink==false)&&(!S_ISREG(mode)))) {
    // Keine regulaere Datei
    return filetype;
  }
  if(size<1) {
    return filetype;
  }
#ifdef HAVE_LIBZ
  #define fopen gzopen
  #define fclose gzclose
  #define fread(B,N,L,F) gzread(F,B,(L)*(N))
#endif
  fp=fopen(fullname,"r");
  if(fp!=NULL) {
    reads=fread(fbuffer,1,64,fp);
    fclose(fp);
    id=ftlist->initEnum();
    ft=(WCFiletype*)ftlist->getFirstElement(id);
    entry=0;
    found=-1;
    WCFiletype *foundft=NULL;
    short *filedesc;
    while(ft!=NULL) {
      if(ft->getinternID()==NORMALTYPE) {
        if(ft->getUseFiledesc()==true) {
          filedesc=ft->getFiledesc();
          for(x=0;x<64;x++) {
            if(filedesc[x]>-1) minsize=x+1;
          }
          if(reads>=minsize) {
            y=0;
            for(x=0;x<reads;x++) {
              if(filedesc[x]>=0) {
                if(filedesc[x]!=(short)fbuffer[x]) {
                  y=1;
                  break;
                }
              }
            }
            if(y==0) {
              foundft=ft;
              break;
            }
          }
        }
        if(ft->getUsePattern()==true) {
          if(match(ft->getPattern())==true) {
            foundft=ft;
            break;
          }
        }
      }
      ft=(WCFiletype*)ftlist->getNextElement(id);
    }
    if(foundft==NULL) {
      filetype=unknownft;
    } else {
      filetype=foundft;
    }
    ftlist->closeEnum(id);
  }
//printf("end check\n");
  return filetype;
#ifdef HAVE_LIBZ
  #undef fopen
  #undef fclose
  #undef fread
#endif
}

bool FileEntry::match(char* pattern)
{
  if(fnmatch(pattern,name,0)==0) return true;
  return false;
}

int Datei::overreadChunk()
{
  int chunksize=getInt();
  while(chunksize>0) {
    getUChar();
    chunksize--;
  }
  return 0;
}

int Datei::getIntSize()
{
  return 5;
}

int Datei::getUCharSize()
{
  return 1;
}

int Datei::getUShortSize()
{
  return 2;
}

int Datei::getULongSize()
{
  return 4;
}

long Datei::errors()
{
  long e=error;
  error=0;
  return e;
}

char *ParentDir(const char *pathstr,char *dirnamereturn,int returnsize)
{
  char *newpath;
  int x,y;
  x=strlen(pathstr);
  if(x<2) {
    newpath=(char*)_allocsafe(2);
    strcpy(newpath,"/");
    if(dirnamereturn!=NULL) strcpy(dirnamereturn,"");
  } else {
    if(pathstr[x-1]=='/') y=x-2; else y=x-1;
    while((pathstr[y]!='/')&&(y>0)) {
      y--;
    }
    if(y==0) {
      newpath=(char*)_allocsafe(2);
      strcpy(newpath,"/");
      if(dirnamereturn!=NULL) {
        strncpy(dirnamereturn,pathstr+1,returnsize);
        dirnamereturn[returnsize-1]=0;
      }
    } else {
      newpath=(char*)_allocsafe(y+1);
      strncpy(newpath,pathstr,y);
      if(dirnamereturn!=NULL) {
        strncpy(dirnamereturn,pathstr+y+1,returnsize);
        dirnamereturn[returnsize-1]=0;
      }
      newpath[y]=0;
    }
  }
  return newpath;
}

int Verzeichnis::getUseSize()
{
  int size=0;
  if(files!=NULL) {
    int id=files->initEnum();
    FileEntry *fe=(FileEntry*)files->getFirstElement(id);
    while(fe!=NULL) {
      if(fe->use==true) size++;
      fe=(FileEntry*)files->getNextElement(id);
    }
    files->closeEnum(id);
  }
  return size;
}

char *Datei::getFilenameFromPath(const char*path)
{
  char *tstr;
  int pos;
  if(path==NULL) return NULL;
  pos=strlen(path)-2;  // -2 because a name contains min. 1 character
                       // but it can be a slash so skiping it
  while(pos>=0) {
    if(path[pos]=='/') break;
    pos--;
  }
  if(pos>=0) tstr=dupstring(&path[pos+1]);
  else tstr=dupstring(path);
  return tstr;
}

char *Datei::getNameWithoutExt(const char *name)
{
  char *tstr;
  int pos,len;
  bool found=false;
  if(name==NULL) return NULL;
  len=strlen(name);
  pos=len-1;
  while(pos>=0) {
    if((name[pos]=='/')&&(pos<(len-1))) break;
    if(name[pos]=='.') {
      found=true;
      break;
    }
    pos--;
  }
  if(found==false) tstr=dupstring(name);
  else {
    tstr=(char*)_allocsafe(pos+1);
    if(pos>0) strncpy(tstr,name,pos);
    tstr[pos]=0;
  }
  return tstr;
}

char *Datei::createTMPName()
{
  char *returnstr,buffer[128];
  int rannr;
  
  for(;;) {
    rannr=rand();
    sprintf(buffer,"%d",rannr);
    returnstr=(char*)_allocsafe(strlen("/tmp/worker")+strlen(buffer)+1);
    sprintf(returnstr,"/tmp/worker%s",buffer);
    if(Datei::lfileExistsExt(returnstr)==Datei::D_FE_NOFILE) break;
    _freesafe(returnstr);
  }
  return returnstr;
}

int Datei::putStringExt(const char *format,const char *str)
{
  if(fp==NULL) return -1;
  fprintf(fp,format,str);
  return 0;
}

char *HandlePathExt(const char *src)
{
  char *tstr,*tstr2;
  char *newpath;
  char *buf;
  int size=4096,i,j;

  newpath=dupstring("");
  int pos=0,count;

  // resolve "."
  if(src[0]=='.') {
    if((src[1]==0)||(src[1]=='/')) {
      for(;;) {
        buf=(char*)_allocsafe(size);
        if(getcwd(buf,size)!=NULL) break;
        _freesafe(buf);
        size*=2;
      }
      _freesafe(newpath);
      if(src[1]=='/') newpath=catstring(buf,"/");
      else newpath=dupstring(buf);
      _freesafe(buf);
      pos=2;
    }
  }
  for(i=pos;i<(int)strlen(src);i++) {
    if(src[i]=='$') {
      // there is an env-variable to resolve
      if(i>pos) {
        // copy the chars before
        buf=(char*)_allocsafe(i-pos+1);
        strncpy(buf,src+pos,i-pos);
        buf[i-pos]=0;
        tstr=catstring(newpath,buf);
        _freesafe(newpath);
        _freesafe(buf);
        newpath=tstr;
      }
      // now find the end of the env-variable
      if(src[i+1]=='{') {
        count=1;
        // search for closing bracket
        for(j=i+2;;j++) {
          if(src[j]=='{') count++;
          else if(src[j]==0) break;
          else if(src[j]=='}') count--;
          if(count==0) {
            j++; // to point at the first char after this
            break;
          }
        }
        // j-1 is the closing bracket
        if((j-2)>=(i+2)) {
          tstr=dupstring(src+i+2);
          tstr[(j-2)-(i+2)+1]=0;
          tstr2=resolveEnv(tstr);
          _freesafe(tstr);
          tstr=catstring(newpath,tstr2);
          _freesafe(newpath);
          _freesafe(tstr2);
          newpath=tstr;
        }
      } else {
        // find the "/" or the end
        for(j=i+1;;j++) {
          if(src[j]==0) break;
          else if(src[j]=='/') break;
        }
        // j-1 is last char for the env
        if((j-1)>=(i+1)) {
          tstr=dupstring(src+i+1);
          tstr[(j-1)-(i+1)+1]=0;
          tstr2=resolveEnv(tstr);
          _freesafe(tstr);
          tstr=catstring(newpath,tstr2);
          _freesafe(newpath);
          _freesafe(tstr2);
          newpath=tstr;
        }
      }
      pos=j;
      i=pos;
    }
  }
  if(i>pos) {
    // copy the chars before
    buf=(char*)_allocsafe(i-pos+1);
    strncpy(buf,src+pos,i-pos);
    buf[i-pos]=0;
    tstr=catstring(newpath,buf);
    _freesafe(newpath);
    _freesafe(buf);
    newpath=tstr;
  }
  // now resolve the normal things like ".." "." and so on
  tstr=HandlePath(newpath);
  _freesafe(newpath);
  return tstr;
}

char *resolveEnv(const char* str)
{
  char *resstr,*tstr,*tstr2;
  char *buf;
  int i,j;

  resstr=dupstring("");
  int pos=0,count;

  for(i=pos;i<(int)strlen(str);i++) {
    if(str[i]=='$') {
      // there is an env-variable to resolve
      if(i>pos) {
        // copy the chars before
        buf=(char*)_allocsafe(i-pos+1);
        strncpy(buf,str+pos,i-pos);
        buf[i-pos]=0;
        tstr=catstring(resstr,buf);
        _freesafe(resstr);
        _freesafe(buf);
        resstr=tstr;
      }
      // now find the end of the env-variable
      if(str[i+1]=='{') {
        count=1;
        // search for closing bracket
        for(j=i+2;;j++) {
          if(str[j]=='{') count++;
          else if(str[j]==0) break;
          else if(str[j]=='}') count--;
          if(count==0) {
            j++; // to point at the first char after this
            break;
          }
        }
        // j-1 is the closing bracket
        if((j-2)>=(i+2)) {
          tstr=dupstring(str+i+2);
          tstr[(j-2)-(i+2)+1]=0;
          tstr2=resolveEnv(tstr);
          _freesafe(tstr);
          tstr=catstring(resstr,tstr2);
          _freesafe(resstr);
          _freesafe(tstr2);
          resstr=tstr;
        }
      } else {
        // find the "/" or the end
        for(j=i+1;;j++) {
          if(str[j]==0) break;
          else if(str[j]=='/') break;
        }
        // j-1 is last char for the env
        if((j-1)>=(i+1)) {
          tstr=dupstring(str+i+1);
          tstr[(j-1)-(i+1)+1]=0;
          tstr2=resolveEnv(tstr);
          _freesafe(tstr);
          tstr=catstring(resstr,tstr2);
          _freesafe(resstr);
          _freesafe(tstr2);
          resstr=tstr;
        }
      }
      pos=j;
    }
  }
  if(i>pos) {
    // copy the chars before
    buf=(char*)_allocsafe(i-pos+1);
    strncpy(buf,str+pos,i-pos);
    buf[i-pos]=0;
    tstr=catstring(resstr,buf);
    _freesafe(resstr);
    _freesafe(buf);
    resstr=tstr;
  }
  tstr=getenv(resstr);
  _freesafe(resstr);
  if(tstr==NULL) resstr=dupstring("");
  else resstr=dupstring(tstr);
  return resstr;
}

void Verzeichnis::setLVC4FE_DND(ListViewContent *lvc,FileEntry *fe,int *sets,ListViewDND *lv,int pos)
{
  setLVC4FE(lvc,fe,sets);
  if(lv!=NULL) lv->replacedndtext(pos,fe->name);
}

